""" Utility methods

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2021-07-07
:Copyright: 2021, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from .data_model import KISAO_ALGORITHM_MAP, UpdatePolicy  # noqa: F401
from biosimulators_utils.config import Config  # noqa: F401
from biosimulators_utils.report.data_model import VariableResults
from biosimulators_utils.sedml.data_model import (  # noqa: F401
    ModelLanguage, Simulation, UniformTimeCourseSimulation, Symbol)
from biosimulators_utils.simulator.utils import get_algorithm_substitution_policy
from biosimulators_utils.utils.core import validate_str_value, parse_value
from biosimulators_utils.warnings import warn, BioSimulatorsWarning
from kisao.data_model import AlgorithmSubstitutionPolicy, ALGORITHM_SUBSTITUTION_POLICY_LEVELS
from kisao.utils import get_preferred_substitute_algorithm_by_ids
from ginsim.gateway import japi as ginsim_japi
import biolqm  # noqa: F401
import biosimulators_utils.model_lang.sbml.utils
import biosimulators_utils.sedml.validation
import biosimulators_utils.xml.utils
import lxml.etree  # noqa: F401
import numpy
import os
import py4j.java_gateway  # noqa: F401
import types  # noqa: F401

__all__ = [
    'validate_simulation',
    'validate_time_course',
    'get_variable_target_xpath_ids',
    'validate_variables',
    'read_model',
    'set_up_simulation',
    'exec_simulation',
    'get_variable_results',
]


def validate_simulation(simulation):
    """ Validate a simulation

    Args:
        simulation (:obj:`Simulation`): simulation

    Returns:
        :obj:`tuple`:

            * nested :obj:`list` of :obj:`str`: nested list of errors (e.g., required ids missing or ids not unique)
            * nested :obj:`list` of :obj:`str`: nested list of errors (e.g., required ids missing or ids not unique)
    """
    errors = []
    warnings = []

    if isinstance(simulation, UniformTimeCourseSimulation):
        temp_errors, temp_warnings = validate_time_course(simulation)
        errors.extend(temp_errors)
        warnings.extend(temp_warnings)

    return (errors, warnings)


def validate_time_course(simulation):
    """ Validate a time course

    Args:
        simulation (:obj:`UniformTimeCourseSimulation`): simulation

    Returns:
        :obj:`tuple`:

            * nested :obj:`list` of :obj:`str`: nested list of errors (e.g., required ids missing or ids not unique)
            * nested :obj:`list` of :obj:`str`: nested list of errors (e.g., required ids missing or ids not unique)
    """
    errors = []
    warnings = []

    if simulation.initial_time != 0:
        errors.append(['Initial time must be 0, not `{}`.'.format(simulation.initial_time)])

    if simulation.output_start_time != int(simulation.output_start_time):
        errors.append(['Output start time must be an integer, not `{}`.'.format(simulation.output_start_time)])

    if simulation.output_end_time != int(simulation.output_end_time):
        errors.append(['Output end time must be an integer, not `{}`.'.format(simulation.output_end_time)])

    step_size = (simulation.output_end_time - simulation.output_start_time) / simulation.number_of_steps
    if abs(step_size - round(step_size)) > 1e-8:
        msg = (
            'The interval between the output start and time time '
            'must be an integer multiple of the number of steps, not `{}`:'
            '\n  Output start time: {}'
            '\n  Output end time: {}'
            '\n  Number of steps: {}'
        ).format(step_size, simulation.output_start_time, simulation.output_end_time, simulation.number_of_steps)
        errors.append([msg])

    return (errors, warnings)


def get_variable_target_xpath_ids(variables, model_etree):
    """ Get the SBML-qual id for each XML XPath target of a SED-ML variable

    Args:
        variables (:obj:`list` of :obj:`Variable`): variables of data generators
        model_etree (:obj:`lxml.etree._ElementTree`): element tree for model

    Returns:
        :obj:`dict`: dictionary that maps each variable target to the id of the
            corresponding qualitative species
    """
    namespaces = biosimulators_utils.xml.utils.get_namespaces_for_xml_doc(model_etree)
    sbml_qual_prefix, sbml_qual_uri = biosimulators_utils.model_lang.sbml.utils.get_package_namespace('qual', namespaces)
    return biosimulators_utils.sedml.validation.validate_target_xpaths(
        variables,
        model_etree,
        attr={
            'namespace': {
                'prefix': sbml_qual_prefix,
                'uri': sbml_qual_uri,
            },
            'name': 'id',
        }
    )


def validate_variables(variables, model, model_language, target_xpath_ids, simulation):
    """ Get the result of each SED-ML variable

    Args:
        variables (:obj:`list` of :obj:`Variable`): variables
        model (:obj:`py4j.java_gateway.JavaObject`): bioLQM model
        model_language (:obj:`str`): model language
        target_xpath_ids (:obj:`dict`): dictionary that maps XPaths to the SBML qualitative ids
            of the corresponding objects
        simulation (:obj:`Simulation`): analysis
    """
    component_ids = set(component.getNodeID() for component in model.getComponents())
    invalid_variables = []

    for variable in variables:
        if variable.symbol:
            if not (isinstance(simulation, UniformTimeCourseSimulation) and variable.symbol == Symbol.time.value):
                invalid_variables.append('{}: symbol: {}'.format(variable.id, variable.symbol))
        else:
            if model_language == ModelLanguage.SBML.value:
                id = target_xpath_ids[variable.target]
            else:
                id = variable.target

            if id not in component_ids:
                invalid_variables.append('{}: target: {}'.format(variable.id, variable.target))

    if invalid_variables:
        valid_variables = []

        if isinstance(simulation, UniformTimeCourseSimulation):
            valid_variables.append('symbol: {}'.format(Symbol.time.value))

        for component_id in component_ids:
            if model_language == ModelLanguage.SBML.value:
                valid_variables.append(
                    "target: /sbml:sbml/sbml:model/qual:listOfQualitativeSpecies/qual:qualitativeSpecies[@id='{}']".format(component_id))
            else:
                valid_variables.append(
                    'target: {}'.format(component_id))

        raise ValueError((
            'The following variables cannot be recorded:\n'
            '  {}\n'
            '\n'
            'Variables with the following symbols and targets can be recorded:\n'
            '  {}'
        ).format(
            '\n  '.join(sorted(invalid_variables)),
            '\n  '.join(sorted(valid_variables)),
        ))


def read_model(filename, language):
    """ Read a model

    Args:
        language (:obj:`ModelLanguage`): language

    Returns:
        :obj:`py4j.java_gateway.JavaObject`: model
    """
    if language == ModelLanguage.SBML:
        format = 'sbml'
    else:
        format = None

    if not os.path.isfile(filename):
        raise FileNotFoundError('`{}` is not a file.'.format(filename))

    model = ginsim_japi.lqm.load(filename, format)

    if model is None:
        raise ValueError('Model `{}` could not be loaded.'.format(filename))

    return model


def set_up_simulation(simulation, config=None):
    """ Set up an analysis

    Args:
        simulation (:obj:`Simulation`): analysis
        config (:obj:`Config`, optional): configuration

    Returns:
        :obj:`tuple`:

            * :obj:`str`: KiSAO of algorithm to execute
            * :obj:`str`: name of the :obj:`biolqm` simulation/analysis method
            * :obj:`types.LambdaType` of :obj:`Simulation` -> :obj:`list` of :obj:`str`: arguments for simulation method
    """
    # simulation algorithm
    alg_kisao_id = simulation.algorithm.kisao_id
    alg_substitution_policy = get_algorithm_substitution_policy(config=config)
    exec_kisao_id = get_preferred_substitute_algorithm_by_ids(
        alg_kisao_id, KISAO_ALGORITHM_MAP.keys(),
        substitution_policy=alg_substitution_policy)
    alg_props = KISAO_ALGORITHM_MAP[exec_kisao_id]
    if simulation.__class__ != alg_props['simulation_type']:
        raise NotImplementedError('{} simulations cannot be executed with `{}` ({}).'.format(
            simulation.__class__.__name__, exec_kisao_id, alg_props['name']))

    method = alg_props['method']
    method_args = []

    # Apply the algorithm parameter changes specified by `simulation.algorithm.parameter_changes`
    if exec_kisao_id == alg_kisao_id:
        for change in simulation.algorithm.changes:
            param_props = alg_props['parameters'].get(change.kisao_id, None)

            if param_props is None:
                if (
                    ALGORITHM_SUBSTITUTION_POLICY_LEVELS[alg_substitution_policy]
                    > ALGORITHM_SUBSTITUTION_POLICY_LEVELS[AlgorithmSubstitutionPolicy.NONE]
                ):
                    warn('Unsuported algorithm parameter `{}` was ignored.'.format(change.kisao_id), BioSimulatorsWarning)
                else:
                    raise NotImplementedError('Algorithm parameter `{}` is not supported.'.format(change.kisao_id))
            else:
                if validate_str_value(change.new_value, param_props['type']):
                    parsed_value = parse_value(change.new_value, param_props['type'])
                    method_args.extend(param_props['method_args'](parsed_value))

                else:
                    if (
                        ALGORITHM_SUBSTITUTION_POLICY_LEVELS[alg_substitution_policy]
                        > ALGORITHM_SUBSTITUTION_POLICY_LEVELS[AlgorithmSubstitutionPolicy.NONE]
                    ):
                        msg = 'Unsuported algorithm parameter value `{}` of `{}` was ignored. The value must be a {}.'.format(
                            change.new_value, change.kisao_id, param_props['type'].value)
                        warn(msg, BioSimulatorsWarning)
                    else:
                        msg = '`{}` is not a valid value of algorithm parameter `{}`. The value must be a {}.'.format(
                            change.new_value, change.kisao_id, param_props['type'].value)
                        raise ValueError(msg)
    else:
        for change in simulation.algorithm.changes:
            warn('Unsuported algorithm parameter `{}` was ignored.'.format(change.kisao_id), BioSimulatorsWarning)

    # return
    return (
        exec_kisao_id,
        method,
        lambda simulation: alg_props['method_args'](simulation) + method_args,
    )


def exec_simulation(method_name, model, args=None):
    """ Execute a task

    Args:
        method_name (:obj:`str`): name of the :obj:`biolqm` simulation/analysis method
        model (:obj:`py4j.java_gateway.JavaObject`): model
        args (:obj:`list` of :obj:`str`, optional): argument to :obj:`method`

    Returns:
        :obj:`list` of :obj:`dict`: result of :obj:`method` for :obj:`model` and :obj:`args`
    """
    method = getattr(biolqm, method_name)
    if args:
        args_list = [' '.join(args)]
    else:
        args_list = []
    result = method(model, *args_list)
    return list(result)


def get_variable_results(variables, model_language, target_xpath_ids, simulation, raw_results):
    """ Get the result of each SED-ML variable

    Args:
        variables (:obj:`list` of :obj:`Variable`): variables
        model_language (:obj:`str`): model language
        target_xpath_ids (:obj:`dict`): dictionary that maps XPaths to the SBML qualitative ids
            of the corresponding objects
        simulation (:obj:`Simulation`): simulation
        raw_results (:obj:`list` of :obj:`dict`): predicted simulatioin states

    Returns:
        :obj:`VariableResults`: result of each SED-ML variable
    """
    n_states = len(raw_results)
    variable_results = VariableResults()
    for variable in variables:
        variable_results[variable.id] = numpy.full((n_states,), numpy.nan)

    for i_state, state in enumerate(raw_results):
        for variable in variables:
            if variable.symbol:
                variable_results[variable.id][i_state] = i_state

            else:
                if model_language == ModelLanguage.SBML.value:
                    id = target_xpath_ids[variable.target]
                else:
                    id = variable.target

                variable_results[variable.id][i_state] = state[id]

    if isinstance(simulation, UniformTimeCourseSimulation):
        for key in variable_results.keys():
            variable_results[key] = numpy.concatenate((
                variable_results[key],
                numpy.full((int(simulation.output_end_time) + 1 - n_states,), variable_results[key][-1]),
            ))
        for variable in variables:
            if variable.symbol and variable.symbol == Symbol.time.value:
                variable_results[variable.id] = numpy.linspace(
                    int(simulation.initial_time),
                    int(simulation.output_end_time),
                    int(simulation.output_end_time) + 1)

        step_size = round((simulation.output_end_time - simulation.output_start_time) / simulation.number_of_steps)
        for key in variable_results.keys():
            variable_results[key] = variable_results[key][int(simulation.output_start_time)::step_size]

    return variable_results



""" BioSimulators-compliant command-line interface to the
`GINsim <http://ginsim.org>`_ simulation program.

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2021-07-07
:Copyright: 2021, BioSimulators Team
:License: MIT
"""

from . import get_simulator_version
from ._version import __version__
from .core import exec_sedml_docs_in_combine_archive
from biosimulators_utils.simulator.cli import build_cli

App = build_cli('biosimulators-ginsim', __version__,
                'GINsim', get_simulator_version(), 'http://ginsim.org',
                exec_sedml_docs_in_combine_archive)


def main():
    with App() as app:
        app.run()



""" Data model

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2021-07-07
:Copyright: 2021, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from biosimulators_utils.data_model import ValueType
from biosimulators_utils.sedml.data_model import SteadyStateSimulation, UniformTimeCourseSimulation
import collections
import enum

__all__ = [
    'UpdatePolicy',
    'KISAO_ALGORITHM_MAP',
]


class UpdatePolicy(str, enum.Enum):
    """ Update policy """
    synchronous = 'synchronous'
    sequential = 'sequential'
    asynchronous = 'asynchronous'
    complete = 'complete'


KISAO_ALGORITHM_MAP = collections.OrderedDict([
    ('KISAO_0000449', {
        'kisao_id': 'KISAO_0000449',
        'name': 'synchronous logical simulation',
        'simulation_type': UniformTimeCourseSimulation,
        'method': 'trace',
        'method_args': lambda simulation: [
            '-m {:d}'.format(int(simulation.output_end_time)),
            '-u {}'.format(UpdatePolicy.synchronous.value),
        ],
        'parameters': {},
    }),
    ('KISAO_0000657', {
        'kisao_id': 'KISAO_0000657',
        'name': 'sequential logical simulation',
        'simulation_type': UniformTimeCourseSimulation,
        'method': 'trace',
        'method_args': lambda simulation: [
            '-m {:d}'.format(int(simulation.output_end_time)),
            '-u {}'.format(UpdatePolicy.sequential.value),
        ],
        'parameters': {},
    }),
    ('KISAO_0000450', {
        'kisao_id': 'KISAO_0000450',
        'name': 'asynchronous logical simulation',
        'simulation_type': UniformTimeCourseSimulation,
        'method': 'random',
        'method_args': lambda simulation: [
            '-m {:d}'.format(int(simulation.output_end_time)),
            '-u {}'.format(UpdatePolicy.asynchronous.value),
        ],
        'parameters': {},
    }),
    ('KISAO_0000573', {
        'kisao_id': 'KISAO_0000573',
        'name': 'complete logical simulation',
        'simulation_type': UniformTimeCourseSimulation,
        'method': 'random',
        'method_args': lambda simulation: [
            '-m {:d}'.format(int(simulation.output_end_time)),
            '-u {}'.format(UpdatePolicy.complete.value),
        ],
        'parameters': {},
    }),
    ('KISAO_0000659', {
        'kisao_id': 'KISAO_0000659',
        'name': 'Naldi MDD stable state search method',
        'simulation_type': SteadyStateSimulation,
        'method': 'fixpoints',
        'method_args': lambda simulation: [],
        'parameters': {},
    }),
    ('KISAO_0000662', {
        'kisao_id': 'KISAO_0000662',
        'name': 'Klarner ASP trap space identification method',
        'simulation_type': SteadyStateSimulation,
        'method': 'trapspaces',
        'method_args': lambda simulation: [],
        'parameters': {
            'KISAO_0000216': {
                'kisao_id': 'KISAO_0000216',
                'name': 'use reduced model',
                'method_args': lambda value: ['reduced'] if value else [],
                'type': ValueType.boolean,
            },
            # 'KISAO_0000XXX': {
            #    'kisao_id': 'KISAO_0000XXX',
            #    'name': 'get all trapped variables',
            #    'method_args': lambda value: ['all'] if value else [],
            #    'type': ValueType.boolean,
            # },
            # 'KISAO_0000XXX': {
            #    'kisao_id': 'KISAO_0000XXX',
            #    'name': 'only get terminal trap spaces',
            #    'method_args': lambda value: ['terminal'] if value else [],
            #    'type': ValueType.boolean,
            # },
            # 'KISAO_0000XXX': {
            #    'kisao_id': 'KISAO_0000XXX',
            #    'name': 'get trap space hierarchy',
            #    'method_args': lambda value: ['tree'] if value else [],
            #    'type': ValueType.boolean,
            # },
        },
    }),
    ('KISAO_0000663', {
        'kisao_id': 'KISAO_0000663',
        'name': 'BDD trap space identification method',
        'simulation_type': SteadyStateSimulation,
        'method': 'trapspaces',
        'method_args': lambda simulation: ['BDD'],
        'parameters': {
            'KISAO_0000216': {
                'kisao_id': 'KISAO_0000216',
                'name': 'use reduced model',
                'method_args': lambda value: ['reduced'] if value else [],
                'type': ValueType.boolean,
            },
            # 'KISAO_0000XXX': {
            #    'kisao_id': 'KISAO_0000XXX',
            #    'name': 'get all trapped variables',
            #    'method_args': lambda value: ['all'] if value else [],
            #    'type': ValueType.boolean,
            # },
            # 'KISAO_0000XXX': {
            #    'kisao_id': 'KISAO_0000XXX',
            #    'name': 'only get terminal trap spaces',
            #    'method_args': lambda value: ['terminal'] if value else [],
            #    'type': ValueType.boolean,
            # },
            # 'KISAO_0000XXX': {
            #    'kisao_id': 'KISAO_0000XXX',
            #    'name': 'get trap space hierarchy',
            #    'method_args': lambda value: ['tree'] if value else [],
            #    'type': ValueType.boolean,
            # },
        },
    }),
])



""" BioSimulators-compliant command-line interface to the `GINsim <http://ginsim.org/>`_ simulation program.

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2021-07-07
:Copyright: 2021, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from .utils import (validate_simulation, get_variable_target_xpath_ids, validate_variables,
                    read_model, set_up_simulation, exec_simulation, get_variable_results)
from biosimulators_utils.combine.exec import exec_sedml_docs_in_archive
from biosimulators_utils.config import get_config, Config  # noqa: F401
from biosimulators_utils.log.data_model import CombineArchiveLog, TaskLog, StandardOutputErrorCapturerLevel  # noqa: F401
from biosimulators_utils.viz.data_model import VizFormat  # noqa: F401
from biosimulators_utils.report.data_model import ReportFormat, VariableResults, SedDocumentResults  # noqa: F401
from biosimulators_utils.sedml import validation
from biosimulators_utils.sedml.data_model import (Task, ModelLanguage, ModelAttributeChange,  # noqa: F401
                                                  SteadyStateSimulation, UniformTimeCourseSimulation,
                                                  Variable, Symbol)
from biosimulators_utils.sedml.exec import exec_sed_doc as base_exec_sed_doc
from biosimulators_utils.sedml.utils import apply_changes_to_xml_model
from biosimulators_utils.utils.core import raise_errors_warnings
import copy
import lxml.etree
import os
import tempfile

__all__ = ['exec_sedml_docs_in_combine_archive', 'exec_sed_doc', 'exec_sed_task', 'preprocess_sed_task']


def exec_sedml_docs_in_combine_archive(archive_filename, out_dir, config=None):
    """ Execute the SED tasks defined in a COMBINE/OMEX archive and save the outputs

    Args:
        archive_filename (:obj:`str`): path to COMBINE/OMEX archive
        out_dir (:obj:`str`): path to store the outputs of the archive

            * CSV: directory in which to save outputs to files
              ``{ out_dir }/{ relative-path-to-SED-ML-file-within-archive }/{ report.id }.csv``
            * HDF5: directory in which to save a single HDF5 file (``{ out_dir }/reports.h5``),
              with reports at keys ``{ relative-path-to-SED-ML-file-within-archive }/{ report.id }`` within the HDF5 file

        config (:obj:`Config`, optional): BioSimulators common configuration

    Returns:
        :obj:`tuple`:

            * :obj:`SedDocumentResults`: results
            * :obj:`CombineArchiveLog`: log
    """
    return exec_sedml_docs_in_archive(exec_sed_doc, archive_filename, out_dir,
                                      apply_xml_model_changes=True,
                                      config=config)


def exec_sed_doc(doc, working_dir, base_out_path, rel_out_path=None,
                 apply_xml_model_changes=True,
                 log=None, indent=0, pretty_print_modified_xml_models=False,
                 log_level=StandardOutputErrorCapturerLevel.c, config=None):
    """ Execute the tasks specified in a SED document and generate the specified outputs

    Args:
        doc (:obj:`SedDocument` or :obj:`str`): SED document or a path to SED-ML file which defines a SED document
        working_dir (:obj:`str`): working directory of the SED document (path relative to which models are located)

        base_out_path (:obj:`str`): path to store the outputs

            * CSV: directory in which to save outputs to files
              ``{base_out_path}/{rel_out_path}/{report.id}.csv``
            * HDF5: directory in which to save a single HDF5 file (``{base_out_path}/reports.h5``),
              with reports at keys ``{rel_out_path}/{report.id}`` within the HDF5 file

        rel_out_path (:obj:`str`, optional): path relative to :obj:`base_out_path` to store the outputs
        apply_xml_model_changes (:obj:`bool`, optional): if :obj:`True`, apply any model changes specified in the SED-ML file before
            calling :obj:`task_executer`.
        log (:obj:`SedDocumentLog`, optional): log of the document
        indent (:obj:`int`, optional): degree to indent status messages
        pretty_print_modified_xml_models (:obj:`bool`, optional): if :obj:`True`, pretty print modified XML models
        log_level (:obj:`StandardOutputErrorCapturerLevel`, optional): level at which to log output
        config (:obj:`Config`, optional): BioSimulators common configuration
        simulator_config (:obj:`SimulatorConfig`, optional): tellurium configuration

    Returns:
        :obj:`tuple`:

            * :obj:`ReportResults`: results of each report
            * :obj:`SedDocumentLog`: log of the document
    """
    return base_exec_sed_doc(exec_sed_task, doc, working_dir, base_out_path,
                             rel_out_path=rel_out_path,
                             apply_xml_model_changes=apply_xml_model_changes,
                             log=log,
                             indent=indent,
                             pretty_print_modified_xml_models=pretty_print_modified_xml_models,
                             log_level=log_level,
                             config=config)


def exec_sed_task(task, variables, preprocessed_task=None, log=None, config=None):
    """ Execute a task and save its results

    Args:
        task (:obj:`Task`): task
        variables (:obj:`list` of :obj:`Variable`): variables that should be recorded
        preprocessed_task (:obj:`dict`, optional): preprocessed information about the task, including possible
            model changes and variables. This can be used to avoid repeatedly executing the same initialization
            for repeated calls to this method.
        log (:obj:`TaskLog`, optional): log for the task
        config (:obj:`Config`, optional): BioSimulators common configuration

    Returns:
        :obj:`tuple`:

            :obj:`VariableResults`: results of variables
            :obj:`TaskLog`: log

    Raises:
        :obj:`NotImplementedError`:

          * Task requires a time course that GINsim doesn't support
          * Task requires an algorithm that GINsim doesn't support
    """
    config = config or get_config()

    if config.LOG and not log:
        log = TaskLog()

    if preprocessed_task is None:
        preprocessed_task = preprocess_sed_task(task, variables, config=config)

    # validate task
    model = task.model
    sim = task.simulation

    # read model
    biolqm_model = preprocessed_task['model']
    variable_target_sbml_id_map = preprocessed_task['variable_target_sbml_id_map']

    # modify model
    if model.changes:
        if model.language == ModelLanguage.SBML.value:
            allowed_change_types = ()
            # allowed_change_types = (ModelAttributeChange,) # TODO: uncomment when bioLQM recognizes initial state
            raise_errors_warnings(
                validation.validate_model_change_types(model.changes, allowed_change_types),
                error_summary='Changes for model `{}` are not supported.'.format(model.id))

            model_etree = preprocessed_task['model_etree']

            model = copy.deepcopy(model)
            for change in model.changes:
                change.new_value = str(int(float(change.new_value)))

            apply_changes_to_xml_model(model, model_etree, sed_doc=None, working_dir=None)

            model_file, model_filename = tempfile.mkstemp(suffix='.xml')
            os.close(model_file)

            model_etree.write(model_filename,
                              xml_declaration=True,
                              encoding="utf-8",
                              standalone=False,
                              pretty_print=False)

            biolqm_model = read_model(model_filename, model.language)

            os.remove(model_filename)

        else:
            raise_errors_warnings(
                validation.validate_model_change_types(model.changes, ()),
                error_summary='Changes for model `{}` are not supported.'.format(model.id))

    # setup simulation
    raise_errors_warnings(
        *validate_simulation(sim),
        error_summary='Simulation `{}` is invalid.'.format(sim.id))

    alg_method_name = preprocessed_task['algorithm_method_name']
    alg_method_args = preprocessed_task['algorithm_method_args'](sim)

    # run simulation
    raw_results = exec_simulation(alg_method_name, biolqm_model, alg_method_args)

    # transform results
    variable_results = get_variable_results(variables, model.language, variable_target_sbml_id_map, sim, raw_results)

    # log action
    if config.LOG:
        log.algorithm = preprocessed_task['algorithm_kisao_id']
        log.simulator_details = {
            'method': alg_method_name,
            'arguments': alg_method_args,
        }

    ############################
    # return the result of each variable and log
    return variable_results, log


def preprocess_sed_task(task, variables, config=None):
    """ Preprocess a SED task, including its possible model changes and variables. This is useful for avoiding
    repeatedly initializing tasks on repeated calls of :obj:`exec_sed_task`.

    Args:
        task (:obj:`Task`): task
        variables (:obj:`list` of :obj:`Variable`): variables that should be recorded
        config (:obj:`Config`, optional): BioSimulators common configuration

    Returns:
        :obj:`dict`: preprocessed information about the task
    """
    config = config or get_config()

    # validate task
    model = task.model
    sim = task.simulation

    if config.VALIDATE_SEDML:
        raise_errors_warnings(
            validation.validate_task(task),
            error_summary='Task `{}` is invalid.'.format(task.id))
        raise_errors_warnings(
            validation.validate_model_language(
                model.language,
                (ModelLanguage.SBML, ModelLanguage.ZGINML)),
            error_summary='Language for model `{}` is not supported.'.format(model.id))

        if model.language == ModelLanguage.SBML.value:
            allowed_change_types = (ModelAttributeChange,)
        else:
            allowed_change_types = ()
        raise_errors_warnings(
            validation.validate_model_change_types(model.changes, allowed_change_types),
            error_summary='Changes for model `{}` are not supported.'.format(model.id))
        raise_errors_warnings(
            *validation.validate_model_changes(model),
            error_summary='Changes for model `{}` are invalid.'.format(model.id))
        raise_errors_warnings(
            validation.validate_simulation_type(sim,
                                                (UniformTimeCourseSimulation, SteadyStateSimulation)),
            error_summary='{} `{}` is not supported.'.format(sim.__class__.__name__, sim.id))
        raise_errors_warnings(
            *validation.validate_simulation(sim),
            error_summary='Simulation `{}` is invalid.'.format(sim.id))
        raise_errors_warnings(
            *validate_simulation(sim),
            error_summary='Simulation `{}` is invalid.'.format(sim.id))

    if not os.path.isfile(model.source):
        raise FileNotFoundError('`{}` is not a file.'.format(model.source))

    if model.language == ModelLanguage.SBML.value:
        model_etree = lxml.etree.parse(model.source)
        variable_target_sbml_id_map = get_variable_target_xpath_ids(variables, model_etree)
    else:
        model_etree = None
        variable_target_sbml_id_map = None

    # validate model
    if config.VALIDATE_SEDML_MODELS:
        raise_errors_warnings(*validation.validate_model(model, [], working_dir='.'),
                              error_summary='Model `{}` is invalid.'.format(model.id),
                              warning_summary='Model `{}` may be invalid.'.format(model.id))

    # read model
    biolqm_model = read_model(model.source, model.language)

    # validate variables
    validate_variables(variables, biolqm_model, model.language, variable_target_sbml_id_map, sim)

    # setup simulation
    alg_kisao_id, method_name, method_args = set_up_simulation(sim, config=config)

    ############################
    # return preprocessed information
    return {
        'model': biolqm_model,
        'model_etree': model_etree,
        'variable_target_sbml_id_map': variable_target_sbml_id_map,
        'algorithm_kisao_id': alg_kisao_id,
        'algorithm_method_name': method_name,
        'algorithm_method_args': method_args,
    }
